#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>

// Defs
#define DEBUG

#define bbdie(str...) badbar_die(__FILE__, __LINE__, str)
#ifdef DEBUG
	#define bbdebug(lev,str...) badbar_debug(lev, __FILE__, __LINE__, str)
#else
	#define bbdebug(lev,str...)
#endif

// Vars
Display *dpy;            
Window root, win, subwin[256];
XImage *bg;
XImage *sbg[256];

int scrn;
struct tm *tmval;
char color[10]="gray";
char color_bg[10]="darkblue";
int howmuch;
int draw_border, transparent, fix;
int winnum;
int current_parent, current_level;
int tmp_id, tmp_mid, cur_id, cur_mid, tmp_winid;
int debug_level;

struct BButton {
	int id, level, parent; // Id is the button id, level is the level id, parent is the parents id
	int winid; // Setten only if sub_button == 1
	char title[15]; // Title
	char cmd[60]; // Cmd to execute
	char sub_button; // 1 - Is a submenu button , 0 - a regular button
	int beginx, endx; // Cords, Begin and end
};

struct BMenu {
	struct BButton button[256]; // Up to 256 buttons in our Bar.
	int lastid, cur_level, menuid[256];
};

struct BMenu Menu;	

// Some functions fot menu
void menu_init(){
	int i;
	Menu.lastid = 0;
	Menu.cur_level = 0;
	for(i=0; i<256; i++) Menu.menuid[i]=0;
}
	
void addmenu(char title[]){
	Menu.button[Menu.lastid].id = Menu.lastid;
	Menu.button[Menu.lastid].level = Menu.cur_level;
	Menu.button[Menu.lastid].parent = Menu.menuid[Menu.cur_level];
	sprintf(Menu.button[Menu.lastid].title, "%s", title);
	Menu.button[Menu.lastid].sub_button = 1;
	Menu.cur_level++;
	Menu.menuid[Menu.cur_level] = Menu.lastid;
	Menu.lastid++;
}

void addbutton(char title[], char cmd[]){
	Menu.button[Menu.lastid].id = Menu.lastid;
	Menu.button[Menu.lastid].level = Menu.cur_level;
	Menu.button[Menu.lastid].parent = Menu.menuid[Menu.cur_level];
	sprintf(Menu.button[Menu.lastid].title, "%s", title);
	sprintf(Menu.button[Menu.lastid].cmd, "%s", cmd);
	Menu.button[Menu.lastid].sub_button = 0;
	Menu.lastid++;
}

GC gc,gc2,gc3,gc4, subgc[256];
Font font;
XFontStruct* font_;
char *window_name={"BadBar"}; 
int x,y,width,height;
int widths[256];

// Errors
void badbar_die(const char *file, int line, const char* str, ...){
	va_list	ap;
	fprintf(stderr, "FATAL[%s:%d]: ", file, line);
    va_start(ap, str);
    vfprintf(stderr, str, ap);
    va_end(ap);
	fprintf(stderr, " $\n");
	exit(0);
}

void badbar_debug(int level, const char *file, int line, const char* str, ...){
	if(level <= debug_level){	
		va_list	ap;
		fprintf(stderr, "DEBUG[%s:%d]: ", file, line);
	    va_start(ap, str);
   		vfprintf(stderr, str, ap);
   	 	va_end(ap);
		fprintf(stderr, " $\n");
	}
}

// Main drawing funcs
void bb_draw(int lev, int parent){
	int i, ilev;
	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].level == lev && Menu.button[i].parent == parent){
			for(ilev = 0; ilev<lev; ilev++) printf("    ");
			if(Menu.button[i].sub_button) {
				printf("+---[");
				printf("%s]\n", Menu.button[i].title);
				bb_draw(lev+1, Menu.button[i].id);
			}
			else {	
				printf("|- %s {%s}\n", Menu.button[i].title, Menu.button[i].cmd);
			}
		}
	}
}

void hide_all(int id){
	int i;
	for(i=id; i>0; i=get_parent_id(i)){
		bbdebug(1,"unmap window [%d][%d]", i, subwin[Menu.button[i].winid]);
		XSelectInput(dpy, subwin[Menu.button[i].winid], NULL);
		XUnmapWindow(dpy,subwin[Menu.button[i].winid]);
	}
}

void DrawBadBar(Window mwin, GC* mgc, int level, int parent){
	int i, xoff;
	
	if(!transparent)
		XClearWindow(dpy, mwin);

	XMapWindow(dpy, mwin);
	xoff = 2;
	
	if(transparent){
		bbdebug(2,"transparent bg... mwin == win is %s", mwin == win ? "true" : "false");
		if(mwin == win){
			XPutImage(dpy, mwin, mgc, bg, 0,0,0,0,
					  width,height);
			bbdebug(1,"XPutimage mainwin w[%d]", width);
		}else{
			XPutImage(dpy, mwin, mgc, sbg[Menu.button[parent].winid], 0,0,0,0,
					  widths[Menu.button[parent].winid],height);
			bbdebug(1,"Widths[%d]==%d",Menu.button[parent].winid, widths[Menu.button[parent].winid]);
		}
	}

	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].level == level && Menu.button[i].parent == parent){
			XDrawString(dpy, mwin, mgc, xoff+2, 12, Menu.button[i].title, strlen(Menu.button[i].title));
			xoff = Menu.button[i].endx;
			XDrawLine(dpy, mwin, mgc, xoff, 0, xoff, height);
			xoff += 2;
		}
	}
		
	// Draw borders
    if(draw_border) {
    	XDrawLine(dpy, mwin, gc3, 0, 0, xoff, 0);
        XDrawLine(dpy, mwin, gc3, 0, 0, 0, height);      
        XDrawLine(dpy, mwin, gc4, 1, height-1, xoff, height-1);
        XDrawLine(dpy, mwin, gc4, xoff-1, 1, xoff-1, height);  
    }
}

// Background
void GetBackground(int id, int x, int y, int mwidth, int mheight){
	if(id == -1) { // Main win
		bg = XGetImage(dpy, root, x, y, mwidth, mheight, -1, ZPixmap);
		width = mwidth;
		bbdebug(1,"Got bg for main win [%d:%d][%d:%d]",x,y,mwidth, mheight);
	}else{
		bbdebug(1,"Getting bg for winid [%d] {+%d+%d %dx%d}",id, x,y, mwidth, mheight);
		sbg[id] = XGetImage(dpy, root, x, y, mwidth, mheight, -1, ZPixmap);
		widths[id] = mwidth;
		bbdebug(1,"Got bg for win[%d:%d] [%d:%d][%d:%d]",id,subwin[id],x,y,mwidth, mheight);
	}
}

// Create window function
void win_create(int id, int mx, int my, int mwidth, int mheight){
    XSetWindowAttributes attr;
    XSizeHints szhints;
    unsigned long mask;
    
    if(fix){ 
        attr.border_pixel = BlackPixel(dpy,scrn);
        attr.background_pixel = GetColor(color_bg);
        attr.override_redirect = True;                                                                 
        attr.event_mask = NULL;
        mask = CWBackPixel|CWBorderPixel|CWOverrideRedirect;
        if(id == -1)
			win = XCreateWindow(dpy,root,mx,my,mwidth,mheight,0,DefaultDepth(dpy,scrn),
								InputOutput,CopyFromParent,mask,&attr);
		else
			subwin[id] = XCreateWindow(dpy,root,mx,my,mwidth,mheight,0,DefaultDepth(dpy,scrn),
								InputOutput,CopyFromParent,mask,&attr);
        szhints.flags = PPosition | PSize;
        szhints.x = mx;
        szhints.y = my;
        szhints.width = mwidth;
        szhints.height = mheight;
        
		if(id == -1)
  		  	XSetNormalHints(dpy,win,&szhints);
		else
  		  	XSetNormalHints(dpy,subwin[id],&szhints);
    }else{
		if(id == -1)
    		win = XCreateSimpleWindow(dpy, root, mx, my, mwidth, mheight, 0, 0, GetColor(color_bg));
		else
    		subwin[id] = XCreateSimpleWindow(dpy, root, mx, my, mwidth, mheight, 0, 0, GetColor(color_bg));
	}
}

// Initing windows and calculating funcs
void init_main_window(){
	// Main window, allways visible
	width = calculate_window_width(0,0);
	win_create(-1, x,y,width,height);
	
	XStoreName(dpy,win,window_name);
	gc = XCreateGC(dpy,win,0,NULL);
	gc2 = XCreateGC(dpy,win,0,NULL);
	gc3 = XCreateGC(dpy,win,0,NULL);
	gc4 = XCreateGC(dpy,win,0,NULL);
	
	XSetForeground(dpy,gc,GetColor(color));
	XSetBackground(dpy,gc,GetColor(color_bg));
	XSetForeground(dpy,gc2,GetColor(color_bg));
	XSetBackground(dpy,gc2,GetColor(color));
	XSetForeground(dpy,gc3,MakeBorderColor(color_bg,1));
	XSetForeground(dpy,gc4,MakeBorderColor(color_bg,2));

  	XSetFont(dpy, gc, font);
  	XSetFont(dpy, gc2, font);

	// Transparent?
	if(transparent)
		GetBackground(-1, x, y, width, height);
}

int calculate_window_width(int lev, int parent){
	int i, ret_width;
	ret_width = width;

	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].level == lev && Menu.button[i].parent == parent){
			ret_width = Menu.button[i].endx;
			bbdebug(2,"    Find width -> num %d | %d", i, ret_width);
		}
	}
	return ret_width;
}

int get_parent_id(int id){
	return Menu.button[id].parent;
}

int get_parent_x_offset(int lev, int id){
	int i,j,lid,nit,ret_x=0;
	
	for(j=lev; j>0; j--){	
		lid = id;
		for(nit=1; nit<j; nit++) lid = get_parent_id(lid);		

		for(i=0; i<Menu.lastid; i++){
			if(Menu.button[i].id == lid){
				ret_x += Menu.button[i].beginx;
				bbdebug(2,"Id [%d:%d], x %d", lid,lev, ret_x);
			}
		}
	}
	return ret_x;
}

int get_id_by_x(int x, int level, int parent){
	int i;
	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].parent == parent && Menu.button[i].level == level 
		   && Menu.button[i].beginx < x && Menu.button[i].endx > x){
			bbdebug(4,"Gotcha! on level %d from parent %d id %d has %d between x1->x2\n",level, parent, i, x);
			return i;
		}
	}
	return -1;
}

int get_id_by_win(Window mwin){
	int i;
	for(i=0; i<Menu.lastid; i++){
		if(subwin[Menu.button[i].winid] == mwin){
			bbdebug(2,"id %d is using window %d",i,mwin);
			return i;
		}
	}
	return -1;
}


void calculate_windows(int parent, int lev){ // Will begin at 0
	int i, xoff;
	xoff = 0;

	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].level == lev && Menu.button[i].parent == parent){
			Menu.button[i].beginx = xoff;
			xoff += XTextWidth(font_, Menu.button[i].title, strlen(Menu.button[i].title)) + 8;
			Menu.button[i].endx = xoff;
			bbdebug(2,"Button id[%d] bx[%d] ex[%d] on level %d", i, Menu.button[i].beginx, Menu.button[i].endx, lev);
			
			// If sub menu -> go further
			if(Menu.button[i].sub_button){	
				calculate_windows(Menu.button[i].id, lev+1);
			}
		}
	}
}

void init_windows(int lev, int parent){
	// All subs	
	int i, mwidth;
	for(i=0; i<Menu.lastid; i++){
		if(Menu.button[i].level == lev && Menu.button[i].parent == parent){
			Menu.button[i].winid = -1;
			if(Menu.button[i].sub_button) {
				// Get width for subwin
				mwidth = calculate_window_width(lev+1, Menu.button[i].id);
				bbdebug(2,"Got width of [%d][%d] = [%d]",Menu.button[i].id, lev+1, mwidth);
				// Create window for this one
				win_create(winnum, x+(get_parent_x_offset(lev+1,Menu.button[i].id)),y+((lev+1)*(height)),mwidth,height);
				bbdebug(1,"Subwindow [%d] is created for [%s][%d] sub", winnum, Menu.button[i].title, Menu.button[i].id);
				Menu.button[i].winid = winnum;
				// Crate GC an set font
				subgc[winnum] = XCreateGC(dpy, subwin[winnum], 0, NULL);
				XSetForeground(dpy, subgc[winnum], GetColor(color));
				XSetBackground(dpy, subgc[winnum], GetColor(color_bg));
				XSetFont(dpy, subgc[winnum], font);
				if(transparent)	
					GetBackground(winnum, x+(get_parent_x_offset(lev+1,Menu.button[i].id)),
								  y+((lev+1)*(height)),mwidth,height);

				winnum++;
				init_windows(lev+1, Menu.button[i].id);
			}
		}
	}
}

// Config
void read_conf(char file[]){
	FILE *fp;
	char line[1024], title[25], val[40];
	char *endl, *p, *titp, *valp;
	int ln;
		
	fp = fopen(file, "r"); 
	if(!fp)
		bbdie("Can't open config file");

	ln = 0;
	while(fgets(line, sizeof(line)-1, fp)){
		// Chomp
		endl = strchr(line, '\n');
		if(endl) *endl = '\0';
	
		// Don't check comments and empty lines
		if(!*line || *line == '#' || *line == '\n'){ ln++; continue; }
		
		// Find begin
		p = &line;
		while(*p == ' ' || *p == '\t')
			p++;

		// Chomp
		int i;
		for(i=strlen(p)-1; i>0; i--){
			if(p[i] == ' ' || p[i] == '\t') p[i] = '\0';
			else break;
		}
	
		// Some Debug
		bbdebug(1, "Config line [%s]",p);	

		if(*p == '[') { // subMenu {beg/end}
			// if begin new submenu
			if(!strcmp(p, "[end]") || !strcmp(p, "[END]")){
				Menu.cur_level--; // Close menu
			}else{
				p++;
				endl = strchr(p, ']');
				if(*endl) *endl = '\0';

				addmenu(p);
			}
		}
		else { // Button
			// Get the title
			int i = 0;
			while(*p != '='){
				title[i] = *p;
				p++; i++;
			}	
			title[i] = '\0';
			
			// Chomp 
			for(i=strlen(title)-1; i>0; i--){
				if(title[i] == ' ' || title[i] == '\t') title[i]='\0';
				else break; 
			}

			// Get the value	
			valp = strchr(p, '=');
			if(!*valp) bbdie("Syntax is Title = Command");
			while(*valp == ' ' || *valp == '=') valp++;

			i = 0;
			while(*valp != '\0'){
				val[i] = *valp;
				valp++; i++;
			}	
			val[i] = '\0';
			addbutton(title, val);
		}
		ln++;
	}
	fclose(fp);
}


/* Now do this by fork()ing twice so we don't have to worry about SIGCHLDs */
void spawn(const char cmd[]) {
        pid_t pid;

        if (dpy)
                putenv(dpy);
        if (!(pid = fork())) {
                setsid();
                switch (fork()) {
                        case 0: system(cmd);
                        default: _exit(0);
                }
        }
        if (pid > 0)
                wait(NULL);
}

unsigned long GetColor(char *color)
{  
  Colormap cmap;
  XColor c0, c1;
  cmap = DefaultColormap(dpy, 0);
  XAllocNamedColor(dpy, cmap, color, &c1, &c0);
  return(c1.pixel);
}

unsigned long MakeBorderColor(char *color, char fgbg){
	Colormap cmap;
	char bordercolor[8] = {0,};
	XColor c0, c1;
	unsigned short r,g,b;

	cmap = DefaultColormap(dpy, 0);
	XAllocNamedColor(dpy, cmap, color, &c1, &c0);
	r = c1.red;
	g = c1.green;
	b = c1.blue;
	
	// If leftupper
	if(fgbg == 1){
		if(r+20000 <= 65535 && g+20000 <= 56635 && b+20000 <= 56635){
			r+=20000; g+=20000; b+=20000;
		}else{
			// Just white
			r = g = b = 56635;
		}
	}
	// If bottomright
	else{
		if(r-40000 >= 0 && g-40000 >= 0 && b-40000 >= 0){
			r-=40000; g-=40000; b-=40000;
		}else{
			// Just black
			r = g = b = 0;
		}
	}

	r = r/256;
	g = g/256;
	b = b/256;
	
	sprintf(bordercolor, "#%lX%lX%lX",r,g,b);
	
	XAllocNamedColor(dpy, cmap, bordercolor, &c1, &c0);
	return c1.pixel;
}

int main(int argc, char **argv){	
	int i, id, curId;
	char cfg[50];	
	char myFont[70];

	XEvent ev;
	XSetWindowAttributes attr;

	draw_border = 1;
	transparent = fix = 0;
	sprintf(cfg,"%s/.badbar2", getenv("HOME"));
	sprintf(myFont,"-misc-fixed-*-*-*-*-18-*-*-*-*-*-*-*");
	width = 200;
	height = 15;
	x = 1;
	y = 1;
	debug_level = 0;

	/* Parse argv */
    for (i=1; i < argc; i++) {
       if (!strcmp(argv[i],"-g") || !strcmp(argv[i],"--geometry")){
          XParseGeometry(argv[i+1],&x,&y,&width,&height);
       }
	   else if (!strcmp(argv[i],"-fg")){
		  sprintf(color,"%s",argv[i+1]);
	   }
	   else if (!strcmp(argv[i],"-bg")){
		  sprintf(color_bg,"%s",argv[i+1]);
	   }	
	   else if (!strcmp(argv[i],"-cfg")){
		  sprintf(cfg,"%s",argv[i+1]);
	   }
	   else if (!strcmp(argv[i],"-f")){
		  sprintf(myFont,"%s",argv[i+1]);
	   }
	   else if (!strcmp(argv[i],"-b")){
		  draw_border = 1;
	   }	
	   else if (!strcmp(argv[i],"+b")){
		  draw_border = 0;
	   }
	   else if (!strcmp(argv[i],"-d")){
		  debug_level = atoi(argv[i+1]);
	   }	   
	   else if (!strcmp(argv[i],"-tr")){
		  transparent = 1;
	   }
       else if (!strcmp(argv[i],"-x")){
		  fix = 1;
	   }
	   else if (!strcmp(argv[i],"-h") || !strcmp(argv[i],"--help")) {
          printf("usage: BadBar [--geometry -x -tr -f[ont] [+-]b[order] -fg -bg -cfg]\n");
          exit(0);
       }
    }

	// Init
	menu_init();
	
	/* Config */
	read_conf(cfg);

	dpy = XOpenDisplay(NULL);
  	scrn = DefaultScreen(dpy);
	root = RootWindow(dpy,scrn);	
	font = XLoadFont(dpy, myFont);
	font_ = XLoadQueryFont(dpy, myFont);
	
	/* Create and Calculate all windows */
	calculate_windows(0,0);
	init_main_window();
	winnum = 0;
	init_windows(0,0);
	
	attr.event_mask = PropertyChangeMask|ButtonPressMask|ButtonReleaseMask|LeaveWindowMask|ExposureMask|PointerMotionMask|EnterWindowMask;
	XChangeWindowAttributes(dpy, win, CWEventMask, &attr);

	XMapWindow(dpy,win);
	XFlush(dpy);
	
	// Draw the ascii menu tree :)
	bb_draw(0,0);
	current_parent = current_level = 0;
	tmp_id = cur_id = 0;
	
	while(1){ 
   		 XNextEvent(dpy, &ev);
         switch (ev.type) {
			 case Expose:
			 case PropertyNotify:
				DrawBadBar(win, gc, 0, 0); // Draw base
				if(cur_id != 0 && current_level != 0 && current_parent == cur_id){	
					DrawBadBar(subwin[Menu.button[cur_id].winid], subgc[Menu.button[cur_id].winid],
					 		   current_level, cur_id);
					bbdebug(1,"Got fucking expose or propnotify [draw win %d]!", subwin[Menu.button[cur_id].winid]);
				}
				break;
			 case LeaveNotify:
				tmp_winid = get_id_by_win(ev.xfocus.window);
				// We just leaved this submenu.
				if(current_level == Menu.button[current_parent].level+1 && tmp_winid == current_parent && tmp_winid != -1){
					bbdebug(1,"leaved window [%d]!", ev.xfocus.window);
					current_level--;
					XSelectInput(dpy, ev.xfocus.window, NULL);
					current_parent = get_parent_id(tmp_winid);
					if(current_level != 0) { 
						XSelectInput(dpy, subwin[Menu.button[current_parent].winid], attr.event_mask); 
					}else{
						XSelectInput(dpy, win, attr.event_mask); 
					}
					XUnmapWindow(dpy, ev.xfocus.window);
				}
				break;
			 case EnterNotify:
					bbdebug(1,"Enterd win [%d]", ev.xfocus.window);
					tmp_winid = get_id_by_win(ev.xfocus.window);
					if(tmp_winid < 0) current_parent = 0; 
					else current_parent = tmp_winid;
					XSelectInput(dpy, ev.xfocus.window, attr.event_mask);
				break;
			 case ButtonPress:
				cur_id = get_id_by_x(ev.xmotion.x, current_level, current_parent);
				if(Menu.button[cur_id].sub_button){
					current_level++;
					XSelectInput(dpy, ev.xmotion.window, ExposureMask);
					current_parent = cur_id;
					bbdebug(3,"DrawBar [%d][%d]", current_level, cur_id);				
					if(cur_id != 0 && current_level != 0){	
						XMapWindow(dpy, subwin[Menu.button[cur_id].winid]);
						DrawBadBar(subwin[Menu.button[cur_id].winid], subgc[Menu.button[cur_id].winid],
					 		   current_level, cur_id);
						XSelectInput(dpy, subwin[Menu.button[cur_id].winid], attr.event_mask);
					}
				}
				else{
					bbdebug(1,"Exec [%s]{%s}",Menu.button[cur_id].title, Menu.button[cur_id].cmd);
					// Hide all windows
					bbdebug(1,"Hideall from parent %d", current_parent);
//					hide_all(current_parent);
					spawn(Menu.button[cur_id].cmd);
//					current_parent = 0;
//					current_level = 0;
//					cur_id = -1;
//					XSelectInput(dpy, win, attr.event_mask);
				}
				break;
			/* case MotionNotify:
					tmp_mid = get_id_by_x(ev.xmotion.x, current_level, current_parent);
					if(cur_mid != tmp_mid){
						cur_mid = tmp_mid;	
						// Hilight :)
						if(current_level != 0){
							XFillRectangle(dpy, subwin[Menu.button[cur_mid].winid], gc, Menu.button[cur_mid].beginx+2,
											2, (Menu.button[cur_mid].endx-Menu.button[cur_mid].beginx)-3, 13);
							XDrawString(dpy, subwin[Menu.button[cur_mid].winid], gc2, Menu.button[cur_mid].beginx+5,
											12, Menu.button[cur_mid].title, strlen(Menu.button[cur_mid].title));
	//						XMapWindow(dpy, subwin[Menu.button[cur_mid].winid]);
						}
						else{
							XFillRectangle(dpy, win, gc, Menu.button[cur_mid].beginx+2,
											2, (Menu.button[cur_mid].endx-Menu.button[cur_mid].beginx)-3, 13);
							XDrawString(dpy, win, gc2, Menu.button[cur_mid].beginx+5,
											12, Menu.button[cur_mid].title, strlen(Menu.button[cur_mid].title));
	//						XMapWindow(dpy, win);
						}
					}
				break;
	*/	}
	}
	return 0;
}
